'use strict';

let constants = require('../constants');
let fs = require('fs');
let crypto = require('crypto');
let async = require('./async');
let request = require('request');
let verbose = require('./verbose');

let sandboxPkey = 'iap-sandbox';
let livePkey = 'iap-live';
let config = null;
let keyPathMap = {};
let publicKeyMap = {};
let googleTokenMap = {};
let checkSubscriptionState = false;
let KEYS = {
	ACCESS_TOKEN: 'access_token',
	GRANT_TYPE: 'grant_type',
	CLIENT_ID: 'client_id',
	CLIENT_SECRET: 'client_secret',
	REFRESH_TOKEN: 'refresh_token'
};
let ENV_PUBLICKEY = {
	SANDBOX: 'GOOGLE_IAB_PUBLICKEY_SANDBOX',
	LIVE: 'GOOGLE_IAB_PUBLICKEY_LIVE'
};
let NAME = '<Google>';

function isValidConfigKey(key) {
	return key.match(/^google/);
}

// test use only
module.exports.reset = function () {
	config = null;
	keyPathMap = {};
	publicKeyMap = {};
	googleTokenMap = {};
	checkSubscriptionState = false;
};

module.exports.readConfig = function (configIn) {
	if (!configIn) {
		// no google iap or public key(s) from ENV variables
		return;
	}
	verbose.setup(configIn);
	config = {};
	let configValueSet = false;
	Object.keys(configIn).forEach(function (key) {
		if (isValidConfigKey(key)) {
			config[key] = configIn[key];
			configValueSet = true;
		}
	});

	// backward compatibility
	if (configIn && configIn.publicKeyStrSandbox) {
		config.googlePublicKeyStrSandbox = configIn.publicKeyStrSandbox;
	}
	if (configIn && configIn.publicKeyStrLive) {
		config.googlePublicKeyStrLive = configIn.publicKeyStrLive;
	}

	if (!configValueSet) {
		config = null;
		return;
	}

	keyPathMap.sandbox = config.googlePublicKeyPath + sandboxPkey;
	keyPathMap.live = config.googlePublicKeyPath + livePkey;

	if (config.googleAccToken && config.googleRefToken && config.googleClientID && config.googleClientSecret) {
		googleTokenMap.accessToken = config.googleAccToken;
		googleTokenMap.refreshToken = config.googleRefToken;
		googleTokenMap.clientID = config.googleClientID;
		googleTokenMap.clientSecret = config.googleClientSecret;
		checkSubscriptionState = true;
	}
};

module.exports.setup = function (cb) {
	if (config && (config.googlePublicKeyStrSandbox || config.googlePublicKeyStrLive)) {
		// try to read public key value as string
		if (config && config.googlePublicKeyStrSandbox) {
			publicKeyMap.sandbox = config.googlePublicKeyStrSandbox;
		}
		if (config && config.googlePublicKeyStrLive) {
			publicKeyMap.live = config.googlePublicKeyStrLive;
		}
		return cb();
	}
	if (!config || !config.googlePublicKeyPath) {
		// try to read public key value from ENV if available
		// if this is set, reading the public key value from file system is ignored
		if (process.env[ENV_PUBLICKEY.SANDBOX]) {
			publicKeyMap.sandbox = process.env[ENV_PUBLICKEY.SANDBOX].replace(/s+$/, '');
		}
		if (process.env[ENV_PUBLICKEY.LIVE]) {
			publicKeyMap.live = process.env[ENV_PUBLICKEY.LIVE].replace(/s+$/, '');
		}
		return cb();
	}
	let keys = Object.keys(keyPathMap);
	async.eachSeries(keys, function (key, next) {
		let pkeyPath = keyPathMap[key];
		fs.readFile(pkeyPath, function (error, fileData) {
			// we are ignoring missing public key file(s)
			if (error) {
				return next();
			}
			publicKeyMap[key] = fileData.toString().replace(/\s+$/, '');
			next();
		});
	}, cb);
};

// receipt is an object
/*
* receipt = { data: 'stringified receipt data', signature: 'receipt signature' };
* if receipt.data is an object, it silently stringifies it
*/
module.exports.validatePurchase = function (dPubkey, receipt, cb) {

	verbose.log(NAME, 'Validate this:', receipt);

	if (typeof receipt !== 'object') {
		verbose.log(NAME, 'Failed: malformed receipt');
		return cb(new Error('malformed receipt: ' + receipt), {
			status: constants.VALIDATION.FAILURE,
			message: 'Malformed receipt'
		});
	}
	if (!receipt.data || !receipt.signature) {
		verbose.log(NAME, 'Failed: missing receipt content');
		return cb(new Error('missing receipt data:\n' + JSON.stringify(receipt)), {
			status: constants.VALIDATION.FAILURE,
			message: 'Malformed receipt'
		});
	}
	if (typeof receipt.data === 'object') {
		// stringify and make sure to escpace the value of developerPayload
		receipt.data = JSON.stringify(receipt.data).replace(/\//g, '\\/');
		verbose.log(NAME, 'Auto stringified receipt data:', receipt.data);
	}

	let pubkey = publicKeyMap.live;

	// override pubkey to allow dynamically fed public key to validate
	if (dPubkey) {
		verbose.log(NAME, 'Using dynamically fed public key:', dPubkey);
		pubkey = dPubkey;
	}

	verbose.log(NAME, 'Try validate against live public key:', pubkey);
	// try live first
	validatePublicKey(receipt, getPublicKey(pubkey), function (error, data) {
		if (error) {
			if (!publicKeyMap.sandbox) {
				verbose.log(NAME, 'Failed to validate against:', pubkey, error);
				return cb(error, {
					status: constants.VALIDATION.FAILURE,
					message: error.message
				});
			}
			pubkey = publicKeyMap.sandbox;
			verbose.log(NAME, 'Failed against live public key:', error);
			verbose.log(NAME, 'Try validate against sandbox public key:', pubkey);
			// now try sandbox
			validatePublicKey(receipt, getPublicKey(pubkey), function (error2, data) {
				if (error2) {
					verbose.log(NAME, 'Failed against sandbox public key:', error2);
					// we will send the error from live only
					return cb(error, {
						status: constants.VALIDATION.FAILURE,
						message: error.message
					});
				}
				verbose.log(NAME, 'Validation against sandbox public key successful:', data);
				// sandbox worked
				checkSubscriptionStatus(data, cb);

			});
			return;
		}

		verbose.log(NAME, 'Validation against live public key successful:', data);
		// live worked
		checkSubscriptionStatus(data, cb);

	});
};

module.exports.getPurchaseData = function (purchase) {
	if (!purchase) {
		return null;
	}
	let data = [];
	let purchaseInfo = {
		transactionId: purchase.purchaseToken,
		orderId: purchase.orderId,
		productId: purchase.productId,
		purchaseDate: purchase.purchaseTime,
		quantity: 1
	};

	if (checkSubscriptionState) {
		purchaseInfo.expirationDate = purchase.expirationTime;
	}

	data.push(purchaseInfo);
	return data;
};

/**
* Function to check subscription status in Google Play
* @param	{Object}	data	receipt data
* @param	{Function}	cb	callback function
*/
function checkSubscriptionStatus(data, cb) {

	data.service = constants.SERVICES.GOOGLE;

	if (!checkSubscriptionState) {
		return cb(null, data);
	}
	let packageName = data.packageName;
	let subscriptionID = data.productId;
	let purchaseToken = data.purchaseToken;

	let url = 'https://www.googleapis.com/androidpublisher/v2/applications/' + packageName +
			'/purchases/subscriptions/' + subscriptionID + '/tokens/' + purchaseToken;
	let state;

	let getSubInfo = function (next) {

		verbose.log(NAME, 'Get subscription info from', url);

		getSubscriptionInfo(url, function (error, response, body) {

			if (error || 'error' in body) {

				verbose.log(NAME, 'Failed to get subscription info from', url, error, body);

				state = constants.VALIDATION.FAILURE;
				// we must move on to validate()
				next();
				return;
			}

			data.autoRenewing = body.autoRenewing;
			data.expirationTime = body.expiryTimeMillis;

			state = constants.VALIDATION.SUCCESS;

			verbose.log(NAME, 'Successfully retrieved subscription info from', url, data);

			next();
		});
	};

	let validate = function (next) {
		switch (state) {
			case constants.VALIDATION.SUCCESS:
				// This line tells the next function there is no need to get subscription Info again.
				// We should read this as a "No, don't call that function again"

				verbose.log(NAME, 'Validated successfully');

				next(null, constants.VALIDATION.FAILURE);
				break;
			case constants.VALIDATION.FAILURE:

				verbose.log(NAME, 'Refresh Google token');

				refreshGoogleTokens(function (error, res, body) {
					if (error) {

						verbose.log(NAME, 'Failed to refresh Google token:', error);

						return cb(error, {
							status: constants.VALIDATION.FAILURE,
							message: error.message
						});
					}

					let parsedBody = JSON.parse(body);

					if ('error' in parsedBody) {

						verbose.log(NAME, 'Failed to refresh Google token:', parsedBody);

						return cb(new Error(parsedBody.error), {
							status: constants.VALIDATION.FAILURE,
							message: parsedBody.error
						});
					}

					// Store new access token
					googleTokenMap.accessToken = parsedBody[KEYS.ACCESS_TOKEN];

					state = constants.VALIDATION.SUCCESS;

					verbose.log(NAME, 'Successfully refreshed Google token:', googleTokenMap.accessToken);

					// On the other hand, here we are telling the next function
					// to get subscription Info again.
					next();
				});
				break;
		}
	};

	let  recheck = function (next) {
		if (state === constants.VALIDATION.SUCCESS) {

			verbose.log(NAME, 'Re-check subscription info:', url);

			getSubscriptionInfo(url, function (error, response, body) {
				if (error || 'error' in body) {

					verbose.log(NAME, 'Re-check failed:', url, error, body);

					state = constants.VALIDATION.FAILURE;
					next(error ? error : new Error(body.error));
					return;
				}
				data.autoRenewing = body.autoRenewing;
				data.expirationTime = body.expiryTimeMillis;
				state = constants.VALIDATION.SUCCESS;

				verbose.log(NAME, 'Re-check successfully retrieved subscription info:', url, data);

				next();
			});
			return;
		}
		// refresh failed
		state = constants.VALIDATION.FAILURE;
		next();
	};

	let done = function (error) {
		if (error) {
			return cb(error, {
				status: constants.VALIDATION.FAILURE,
				message: error.message
			});
		}

		cb(null, data);
	};

	let tasks = [
		getSubInfo,
		validate,
		recheck
	];

	async.series(tasks, done);
}

function getPublicKey(publicKey) {
	if (!publicKey) {
		return null;
	}
	let key = chunkSplit(publicKey, 64, '\n');
	let pkey = '-----BEGIN PUBLIC KEY-----\n' + key + '-----END PUBLIC KEY-----\n';
	return pkey;
}

function validatePublicKey(receipt, pkey, cb) {
	if (!receipt || !receipt.data) {
		return cb(new Error('missing receipt data'));
	}
	if (!pkey) {
		return cb(new Error('missing public key'));
	}
	if (typeof receipt.data !== 'string') {
		return cb(new Error('receipt.data must be a string'));
	}
	let validater = crypto.createVerify('SHA1');
	let valid;
	validater.update(receipt.data);
	try {
		valid = validater.verify(pkey, receipt.signature, 'base64');
	} catch (error) {
		return cb(error);
	}
	if (valid) {
		// validated successfully
		let data = JSON.parse(receipt.data);
		data.status = constants.VALIDATION.SUCCESS;
		return cb(null, data);
	}
	// failed to validate
	cb(new Error('failed to validate purchase'));
}

function chunkSplit(str, len, end) {
	len = parseInt(len, 10) || 76;
	if (len < 1) {
		return false;
	}
	end = end || '\r\n';
	return str.match(new RegExp('.{0,' + len + '}', 'g')).join(end);
}

function getSubscriptionInfo(url, cb) {
	let options = {
		method: 'GET',
		url: url,
		headers: {
			'Authorization': 'Bearer ' + googleTokenMap.accessToken
		},
		json: true
	};

	request(options, cb);
}

module.exports.refreshToken = function (cb) {

	if (!checkSubscriptionState) {
		return cb(new Error('missing google play api info'), {
			status: constants.VALIDATION.FAILURE,
			message: 'client_id, client_secret, access_token and refres_token should be provided'
		});
	}

	refreshGoogleTokens(function (error, res, body) {
		if (error) {
			return cb(error, { status: constants.VALIDATION.FAILURE, message: error.message });
		}

		let parsedBody = JSON.parse(body);

		if ('error' in parsedBody) {
			return cb(new Error(parsedBody.error), {
				status: constants.VALIDATION.FAILURE,
				message: parsedBody.error
			});
		}

		// Store new access token
		googleTokenMap.accessToken = parsedBody[KEYS.ACCESS_TOKEN];
		cb(null, parsedBody);
	});
};

function refreshGoogleTokens(cb) {

	let body = {};
	body[KEYS.GRANT_TYPE] = KEYS.REFRESH_TOKEN;
	body[KEYS.CLIENT_ID] = googleTokenMap.clientID;
	body[KEYS.CLIENT_SECRET] = googleTokenMap.clientSecret;
	body[KEYS.REFRESH_TOKEN] = googleTokenMap.refreshToken;

	let options = {
		method: 'POST',
		url: 'https://accounts.google.com/o/oauth2/token',
		form: body
	};

	request(options, cb);
}
