'use strict';

let verbose = require('./verbose');
let constants = require('../constants');
let xmlCrypto = require('xml-crypto');
let Parser = require('xmldom').DOMParser;
let request = require('request');
let url = 'https://lic.apps.microsoft.com/licensing/certificateserver/?cid=';
let sigXPath = '//*//*[local-name(.)=\'Signature\' and namespace-uri(.)=\'http://www.w3.org/2000/09/xmldsig#\']';

let NAME = '<Windows>';

module.exports.readConfig = function (configIn) {
	if (!configIn) {
		// no required config
		return;
	}
	verbose.setup(configIn);
	// Apply any default settings to Request.
	if ('requestDefaults' in configIn) {
		request = request.defaults(configIn.requestDefaults);
	}
};

// receipt is an XML string... oh microsoft... why?
module.exports.validatePurchase = function (receipt, cb) {
	let certId;
	let options = {
		ignoreWhiteSpace: true,
		errorHandler: {
			fatalError: handleError,
			error: handleError
		}
	};
	let handleError = function (error) {
		if (typeof error === 'string') {
			error = new Error(error);
		}
		cb(error);
	};
	verbose.log(NAME, 'Validate:', receipt);
	try {
		let doc = new Parser(options).parseFromString(receipt);
		certId = doc.firstChild.getAttribute('CertificateId');
	} catch (e) {
		verbose.log(NAME, 'Failed:', e);
		return cb(new Error('failed to validate purchase: ' + e.message), { status: constants.VALIDATION.FAILURE, message: e.message });
	}
	if (!certId) {
		verbose.log(NAME, 'Failed: Invalid certificate ID');
		return cb(new Error('failed to find certificate ID'), { status: constants.VALIDATION.FAILURE, message: 'Invalid certificate ID' });
	}
	verbose.log(NAME, 'Get public key from:', url + certId);
	send(url + certId, function (error, body) {
		if (error) {
			verbose.log(NAME, 'Failed to get public key:', (url + certId), error);
			return cb(error);
		}
		let data;
		try {
			let publicKey = body;
			let canonicalXML = removeWhiteSpace(doc.firstChild).toString();
			let signature = xmlCrypto.xpath(doc, sigXPath);
			let sig = new xmlCrypto.SignedXml();
			sig.keyInfoProvider = new Cert(publicKey);
			sig.loadSignature(signature.toString());
			if (sig.checkSignature(canonicalXML)) {
				// create purchase data
				let items = doc.getElementsByTagName('ProductReceipt');
				let purchases = [];
				for (let i = 0, len = items.length; i < len; i++) {
					let item = items[i];
					purchases.push({
						transactionId: item.getAttribute('Id'),
						productId: item.getAttribute('ProductId'),
						purchaseDate: item.getAttribute('PurchaseDate'),
						expirationDate: item.getAttribute('ExpirationDate'),
						productType: item.getAttribute('ProductType'),
						appId: item.getAttribute('AppId')
					});
				}
				// successful validation
				data = {
					service: constants.SERVICES.WINDOWS,
					status: constants.VALIDATION.SUCCESS,
					purchases: purchases
				};
			}
		} catch (e) {
			verbose.log(NAME, 'Failed to validated:', e);
			return cb(new Error('failed to validate purchase: ' + e.message), { status: constants.VALIDATION.FAILURE, message: e.message });
		}
		// done
		verbose.log(NAME, 'Validation success:', data);
		cb(null, data);
	});
};

module.exports.getPurchaseData = function (purchase, options) {
	if (!purchase || !purchase.purchases || !purchase.purchases.length) {
		return null;
	}
	let data = [];
	for (let i = 0, len = purchase.purchases.length; i < len; i++) {
		let item = purchase.purchases[i];
		let exp = new Date(item.expirationDate).getTime();

		if (options && options.ignoreExpired && exp && Date.now() - exp >= 0) {
			// we are told to ignore expired item and it has been expired
			continue;
		}

		data.push({
			transactionId: item.transactionId,
			productId: item.productId,
			purchaseDate: new Date(item.purchaseDate).getTime(),
			expirationDate: exp,
			quantity: 1
		});
	}
	return data;
};

function send(url, cb) {
	let options = {
		encoding: null,
		url: url
	};
	request.get(options, function (error, res, body) {
		if (error) {
			return cb(error, { status: res.status, message: body });
		}
		if (!body) {
			return cb(new Error('invalid response from the service'), { status: res.status, message: 'Unknown' });
		}
		cb(null, body.toString('utf8'));
	});
}

function removeWhiteSpace(node) {
	let rootNode = node;
	while (node) {
		if (!node.tagName && (node.nextSibling || node.previousSibling)) {
			node.parentNode.removeChild(node);
		}
		removeWhiteSpace(node.firstChild);
		node = node.nextSibling;
	}
	return rootNode;
}

function Cert(pubKey) {
	this._pubKey = pubKey;
}

Cert.prototype.getKeyInfo = function () {
	return '<X509Data></X509Data>';
};

Cert.prototype.getKey = function () {
	return this._pubKey;
};
